#pragma once

#include "SPIFFS.h"
#include "AudioFileSourceSPIFFS.h"
#include "AudioFileSourceID3.h"
#include "AudioGeneratorMP3.h"
#include "AudioOutputI2S.h"

#include "Log.h"

class SpiffPlayer {

private:
  String                _mp3FileName;
  AudioFileSourceSPIFFS *_file;
  AudioOutputI2S        *_out;
  AudioFileSourceID3    *_id3;
  AudioGeneratorMP3     *_mp3;

  bool isSetup;
public:
  SpiffPlayer(String mp3FileName) {
    this->_mp3FileName = mp3FileName;
    this->_mp3 = NULL;
    this->_file = NULL;
    this->_out = NULL;
    this->_id3 = NULL;
  }

  SpiffPlayer() {
    this->isSetup = false;
    this->_mp3FileName = "";
    this->_mp3 = NULL;
    this->_file = NULL;
    this->_out = NULL;
    this->_id3 = NULL;
  }

  ~SpiffPlayer() {
    if (this->_mp3)   delete this->_mp3;
    if (this->_id3)   delete this->_id3;
    if (this->_out)   delete this->_out;
    if (this->_file)  delete this->_file;
  }

private:
  bool checkIfFileExists() {
    LOG_INFO("SpiffPlayer - checkIfFileExists", "looking for mp3 file %s", this->_mp3FileName.c_str());
    if (!SPIFFS.exists(this->_mp3FileName)) {
      LOG_ERROR("SpiffPlayer - checkIfFileExists", "failed to find mp3 file %s", this->_mp3FileName.c_str());
      return false;
    }
    LOG_INFO("SpiffPlayer - checkIfFileExists", "found mp3 file %s", this->_mp3FileName.c_str());
    File mp3File = SPIFFS.open(this->_mp3FileName, "r");
    if (!mp3File) {
      LOG_ERROR("SpiffPlayer - checkIfFileExists", "failed to open %s", this->_mp3FileName.c_str());
      return false;
    }

    LOG_INFO("SpiffPlayer - checkIfFileExists", "mp3 file (%s) has size %u", this->_mp3FileName.c_str(), mp3File.size());
    mp3File.close();

    LOG_INFO("SpiffPlayer - checkIfFileExists", "initializing audio components");
    return true;
  }

  bool setupAudioComponents() {
    LOG_INFO("SpiffPlayer - setupAudioComponents", "setting up audio components");

    this->_file = new AudioFileSourceSPIFFS(this->_mp3FileName.c_str());
    if (!this->_file || !this->_file->isOpen()) {
        LOG_ERROR("SpiffPlayer - setupAudioComponents", "AudioFileSourceSPIFFS could not be opened");
        return false;
    }
    if (!this->_file->getSize()) {
      LOG_WARNING("SpiffPlayer - setupAudioComponents", "audio file is empty");
    }
    LOG_INFO("SpiffPlayer - setupAudioComponents", "Successfully setup AudioFileSourceSPIFFS");

    this->_id3 = new AudioFileSourceID3(this->_file);
    if (!this->_id3) {
      LOG_ERROR("SpiffPlayer - setupAudioComponents", "AudioFileSourceID3 could not be opened");
      return false;
    }
    LOG_INFO("SpiffPlayer - setupAudioComponents", "Successfully setup AudioFileSourceID3");

    this->_out = new AudioOutputI2S();
    if (!this->_out) {
      LOG_ERROR("SpiffPlayer - setupAudioComponents", "AudioOutputI2S could not be initialized");
      return false;
    }
    LOG_INFO("SpiffPlayer - setupAudioComponents", "Successfully setup AudioOutputI2S");

    this->_mp3 = new AudioGeneratorMP3();
    if (!this->_mp3) {
      LOG_ERROR("SpiffPlayer - setupAudioComponents", "AudioGeneratorMP3 could not be initialized");
      return false;
    }
    LOG_INFO("SpiffPlayer - setupAudioComponents", "finished setting up audio components");
    return true;
  }

public:
  bool setup() {
    if (this->isSetup) {
      return true;
    }
    
    if (!SPIFFS.begin(true)) {
      LOG_ERROR("SpiffPlayer - setup", "an error occurred while mounting SPIFFS");
      return false;
    }
    LOG_INFO("SpiffPlayer - setup", "SPIFFS mounted successfully!");

    if (!this->checkIfFileExists()) {
      return false;
    }
    if (!this->setupAudioComponents()) {
      return false;
    }

    LOG_INFO("SpiffPlayer - setup","begin play audio file %s", this->_mp3FileName.c_str());
    if (!this->_mp3->begin(this->_id3, this->_out)) {
      LOG_ERROR("SpiffPlayer - setup", "failed to start mp3 generator");
      return false;
    }
    LOG_INFO("SpiffPlayer - setup", "mp3 playback started.");
    this->isSetup = true;
    return true;
  }

  bool setup(String mp3FileName, bool loop = false) {
    this->_mp3FileName = mp3FileName;
    return this->setup();
  }

  bool stop() {
    if (!this->_mp3) {
      LOG_ERROR("SpiffPlayer - stop", "call setup before using this function");
      return false;
    }
    return this->_mp3->stop();
  }

  bool isRunning() {
    if (!this->_mp3) {
      LOG_ERROR("SpiffPlayer - isRunning", "call setup before using this function");
      return false;
    }
    return this->_mp3->isRunning();
  }

  bool loop() {
    if (!this->_mp3) {
      LOG_ERROR("SpiffPlayer - loop","call setup before using this function");
      return false;
    }
    return this->_mp3->loop();
  }

  // Example of continuous playback if you want looping
  bool continuousPlay(bool stopPlay = false) {
    if (!this->_mp3) {
      LOG_ERROR("SpiffPlayer - continuousPlay", "call setup before using this function");
      return false;
    }
    if (stopPlay) {
      LOG_INFO("SpiffPlayer - continuousPlay", "Stopping continuous playback");
      if (this->_mp3->isRunning()) {
        this->_mp3->stop();
      }
      return false;
    }
    if (this->_mp3->isRunning()) {
      if (!this->_mp3->loop()) {
        LOG_INFO("SpiffPlayer - continuousPlay", "Playback finished, restarting...");
        if (!this->_file->seek(0, SEEK_SET)) {
          LOG_ERROR("SpiffPlayer - continuousPlay", "Failed to reset file pointer");
          return false;
        }
        if (!this->_mp3->begin(this->_id3, this->_out)) {
          LOG_ERROR("SpiffPlayer - continuousPlay", "Failed to start audio generator");
          return false;
        }
      }
    }
    return true;
  }
};
